/*
 *
 *  This file is part of
 *	MakeIndex - A formatter and format independent index processor
 *
 *  Copyright (C) 1989 by Chen & Harrison International Systems, Inc.
 *  Copyright (C) 1988 by Olivetti Research Center
 *  Copyright (C) 1987 by Regents of the University of California
 *
 *  Author:
 *	Pehong Chen
 *	Chen & Harrison International Systems, Inc.
 *	Palo Alto, California
 *	USA
 *	(phc@renoir.berkeley.edu or chen@orc.olivetti.com)
 *
 *  Contributors:
 *	Please refer to the CONTRIB file that comes with this release
 *	for a list of people who have contributed to this and/or previous
 *	release(s) of MakeIndex.
 *
 *  All rights reserved by the copyright holders.  See the copyright
 *  notice distributed with this software for a complete description of
 *  the conditions under which it is made available.
 *
 */

/***********************************************************************

			   INSTALLATION NOTES
			      [08-Apr-1994]

   At compile-time (or here, if compile-time definition is not
   available), set non-zero ONE OS_xxxx operating system flag, and if
   more than one compiler is available and it is necessary to
   distinguish between the compilers, then, and only then, ONE
   compiler flag:

    Operating Systems	     Compiler(s)
    =================	     ===========
    OS_AMIGA
    OS_ATARI
    OS_BS2000
    OS_BSD
    OS_MVSXA
    OS_PCDOS		     IBM_PC_MICROSOFT or IBM_PC_TURBO
    OS_SYSV
    OS_TOPS20		     KCC_20 or PCC_20
    OS_VAXVMS
    OS_VMCMS
    OS_XENIX
    =================	     ===========

   If no OS_xxxx symbol is defined, OS_BSD is assumed.

   If Standard C prototypes are supported, define the symbol
   STDC_PROTOTYPES in the appropriate OS_xxxx section below, and
   insert #include's for the standard system files which define
   library prototypes. STDC_PROTOTYPES will be defined automatically
   if __STDC__ is; the latter must be defined by all Standard C
   conformant implementations.

   All function declarations in MakeIndex are contained at the end of
   this file.  If STDC_PROTOTYPES is not selected, then all the
   standard library functions must be declared explicitly.

   If the host system restricts external names to 6 characters, define
   SHORTNAMES in the appropriate OS_xxxx section below.

   Installing MakeIndex under BS2000 requires at least Version 8.5
   supporting Multiple Public Volume Sets (MPVS). Define WORK in this
   file to your CATID (default :w:) for temporary files (.ilg, .ind),
   in case of no MPVS support define WORK to NIL.

***********************************************************************/

#if (__STDC__ || defined(__cplusplus))
#define STDC		1
#else
#define STDC		0
#endif

/*
 * Establish needed operating symbols (defaulting to OS_BSD if none
 * specified at compile time).	If you add one, add it to the check
 * list at the end of this section, too.
 */

#ifndef    OS_AMIGA
#define OS_AMIGA	0
#endif

#ifndef	   OS_ATARI
#define OS_ATARI	0
#endif

#ifndef	   OS_BS2000
#define OS_BS2000	0
#endif

#ifndef	   OS_BSD
#define OS_BSD		0
#endif

#ifndef	   OS_MVSXA
#define OS_MVSXA	0
#endif

#ifndef	   OS_PCDOS
#define OS_PCDOS	0
#endif

#ifndef	   IBM_PC_MICROSOFT
#define IBM_PC_MICROSOFT 0
#endif

#ifndef	   IBM_PC_TURBO
#define IBM_PC_TURBO	0
#endif

#ifdef __TURBOC__
#define IBM_PC_TURBO	1
#endif

#ifndef	   OS_SYSV
#define OS_SYSV		0
#endif

#ifndef	   OS_TOPS20
#define OS_TOPS20	0
#endif

#ifndef	   KCC_20
#define KCC_20 		0
#endif

#ifndef	   PCC_20
#define PCC_20 		0
#endif

#if    OS_TOPS20
#if    (KCC_20 || PCC_20)
#else
#undef PCC_20
#define PCC_20		1		/* PCC-20 is default for TOPS-20 */
#endif /* KCC_20 || PCC_20) */
#endif /* OS_TOPS20 */

#ifndef	   OS_VAXVMS
#define OS_VAXVMS	0
#endif

#ifndef	   OS_VMCMS
#define OS_VMCMS	0
#endif

#ifndef	   OS_XENIX
#define OS_XENIX	0
#endif

#if	(OS_ATARI || OS_BSD || OS_BS2000 || OS_MVSXA || OS_PCDOS || OS_VMCMS)
#else
#if	(OS_SYSV || OS_TOPS20 || OS_VAXVMS || OS_XENIX || OS_AMIGA)
#else
#undef OS_BSD
#define OS_BSD		1	/* Berkeley UNIX is default operating system */
#endif
#endif

/**********************************************************************/

#define EXIT(code)	exit(code)

#define DIGITVALUE(c)	((page_number)((int)(c) - (int)('0')))
			/* this relation is true in all char sets */
			/* provided c is a digit */

#define ISDIGIT(C)	('0' <= (C) && (C) <= '9')

#define ISSYMBOL(C)	(('!' <= (C) && (C) <= '@') || \
			 ('[' <= (C) && (C) <= '`') || \
			 ('{' <= (C) && (C) <= '~'))

#define OPTION_CHAR(c)	(c)		/* mapping of command-line chars */

#define STATIC				/* empty definition => automatic */

#define STDC_PROTOTYPES	STDC

#define STRCMP(a,b)	strcmp((char*)(a),(char*)(b)) /* used in sortid.c */

#define TOASCII(i)	((char)((i) + (int)('0'))) /* true in all char sets */

#if defined(__BSD_4_4__) && defined(__cplusplus)
/* HP 9000/425s BSD 4.3: we need this BEFORE stdio.h, because signed is
also used in system files included by stdio.h. */
#define signed                  /* signed not implemented in BSD C++ */
				/* but char is a signed type */
#endif


#include <stdio.h>
#include <ctype.h>

#if __alpha && unix
#include <sys/access.h>			/* for R_OK et al */
#endif /* __alpha && unix */

#if    ardent
#include <string.h>
#endif /* ardent */

#if    OS_AMIGA
#include <stddef.h>			/* for function declarations */
#define DIR_DELIM	'/'
#define ENV_SEPAR	','
#define VOL_DELIM	':'
#undef STATIC
#define STATIC		static		/* to conserve stack space */
#undef STDC_PROTOTYPES
#define STDC_PROTOTYPES    1		/* for full argument type checking*/
#define MISSING_UNISTD_H
#endif /* OS_AMIGA */

#if    OS_ATARI
#define DIR_DELIM	'\\'
#define ENV_SEPAR	';'
#define strchr		index		/* convert STDC form to old K&R form */
#define strrchr		rindex		/* convert STDC form to old K&R form */

#if    ATARI_ST_TURBO
#define access(fn, mode) 0		/* function not available */
#endif

#if    __GNUC__
#include <types.h>
#endif

#undef STDC_PROTOTYPES
#define STDC_PROTOTYPES	1		/* so we get full argument checking */
#endif /* OS_ATARI */

#if    OS_BS2000
#include <string.h>
#define ASCII		1
#define DIR_DELIM	'$'
#define ENV_SEPAR	'#'
#define getenv(P)	0		/* function not available */
#undef ISDIGIT
#define ISDIGIT(C)	isdigit(C)
#undef ISSYMBOL
#define ISSYMBOL(C) (! (isdigit(C) || isalpha(C) || iscntrl(C) || (C) == SPC))
#undef OPTION_CHAR
#define OPTION_CHAR(c)	tolower(c)	/* remap upper case to lower case */
#define SHORTNAMES
#undef STRCMP
#define STRCMP(a,b)	new_strcmp((unsigned char*)(a),(unsigned char*)(b), \
				   ASCII) /* used in sortid.c */
#endif /* OS_BS2000 */

#if    OS_BSD
#define DIR_DELIM	'/'
#define ENV_SEPAR	':'
#define strchr		index		/* convert STDC form to old K&R form*/
#define strrchr		rindex		/* convert STDC form to old K&R form */
#endif

#if    OS_MVSXA
#include <stdefs.h>			/* for function declarations */
#define access(fn, mode) 0		/* function not available */
#define ASCII		1
#define DIR_DELIM	'.'
#define ENV_SEPAR	'.'
#undef ISDIGIT
#define ISDIGIT(C)	isdigit(C)
#undef ISSYMBOL
#define ISSYMBOL(C) (! (isdigit(C) || isalpha(C) || iscntrl(C) || (C) == SPC))
#undef OPTION_CHAR
#define OPTION_CHAR(c)	tolower(c)	/* remap upper case to lower case */
#define SHORTNAMES
#undef STDC_PROTOTYPES
#define STDC_PROTOTYPES	1		/* for full argument type checking */
#undef STRCMP
#define STRCMP(a,b)	new_strcmp((unsigned char*)(a),(unsigned char*)(b), \
				   ASCII) /* used in sortid.c */
#endif /* OS_MVSXA */

#if    OS_PCDOS

#if    (IBM_PC_MICROSOFT || IBM_PC_TURBO)
#include <io.h>				/* for function declarations */
#include <stdlib.h>
#include <string.h>
#endif /* (IBM_PC_MICROSOFT || IBM_PC_TURBO) */

#include <stddef.h>			/* for function declarations */
#define DIR_DELIM	'\\'
#define ENV_SEPAR	';'
#undef STATIC
#define STATIC		static		/* to conserve stack space */
#undef STDC_PROTOTYPES
#define STDC_PROTOTYPES	   1		/* for full argument type checking */
#define OPEN_IN_B(FP) fopen(FP, "rb")
#define MISSING_UNISTD_H
#endif /* OS_PCDOS */

#if    OS_SYSV
#define DIR_DELIM	'/'
#define ENV_SEPAR	':'
#endif

#if    OS_TOPS20
#define DIR_DELIM	':'
#define ENV_SEPAR	','
#define SHORTNAMES

#if    PCC_20
#define strchr		index		/* convert STDC form to old K&R form */
#define strrchr		rindex		/* convert STDC form to old K&R form */

#define R_OK		0		/* PCC-20 access(file,mode) */
					/* only understands mode=0  */
#endif /* PCC_20 */
#endif /* OS_TOPS20 */

#if    OS_VAXVMS
#define DIR_DELIM	':'
#define ENV_SEPAR	','		/* In fact these aren't used */
#undef EXIT
#define EXIT(code)	exit( ((code) == 0) ? 1 : (1 << 28) + 2 )
#define INDEX_LOG	".lis"
#define OPTION_CHAR(c)	tolower(c)	/* remap upper case to lower case */
#endif /* OS_VAXVMS */

#if    OS_VMCMS
#include <stdefs.h>			/* for function declarations */
#define access(fn, mode) 0		/* function not available */
#define DIR_DELIM	'.'
#define ENV_SEPAR	'.'
#define SHORTNAMES
#undef STDC_PROTOTYPES
#define STDC_PROTOTYPES	1		/* for full argument type checking */
#endif /* OS_VMCMS */

#if    OS_XENIX
#define ENV_SEPAR	':'
#define DIR_DELIM	'/'
#define strchr		index		/* convert STDC form to old K&R form */
#define strrchr		rindex		/* convert STDC form to old K&R form */
#endif /* OS_XENIX */

#if    STDC_PROTOTYPES
#define ARGS(arg_list)	arg_list
#define VOIDP		void*
#define VOID_ARG	void
#else
#define ARGS(arg_list)	()
#undef const
#define const
#define VOIDP		char*
#define VOID_ARG
#endif

#if defined(__host_mips) && defined(_SYSTYPE_SYSV) && !STDC
#if !(defined(_SIZE_T) || defined(_SIZE_T_))
#define _SIZE_T				/* for MIPS 6280 */
#define _SIZE_T_			/* for SGI IRIX 4.0.x */
typedef unsigned  int    size_t;
#endif
#endif

#if defined(sun) && !STDC
#include <stdlib.h>			/* for size_t */
#endif

#define FREE(p) free((void*)(p))

#if defined(__hppa) && defined(__cplusplus)
#define signed                  /* signed not implemented in HP-UX 9.0 C++ */
				/* but char is a signed type */
#endif


#if defined(sun) && defined(__cplusplus)
#define signed			/* signed not implemented in SunOS 4.1.3 C++ */
				/* but char is a signed type */
#undef FREE
#define FREE(p) free((char*)(p))	/* workaround for arg type error */
#endif

#if __NeXT__
#if __cplusplus
extern "C" int	access ARGS((const char *, int));
#else
int	access ARGS((const char *, int));
#endif
#define MISSING_UNISTD_H
#endif

#if    STDC
#include <stdlib.h>
#undef strchr
#undef strrchr
#include <string.h>

#if !defined(MISSING_UNISTD_H)
#include <unistd.h>
#endif

#else /* NOT STDC */
	/* Miscellaneous standard library routines */

#if !defined(access)
int	access ARGS((const char *, int));
#endif

VOIDP	calloc ARGS((size_t nitems,size_t size));

#ifndef getenv
char   *getenv ARGS((const char *name));
#endif

VOIDP	malloc ARGS((size_t size));
char   *strchr ARGS((const char *s,int c));
char   *strrchr ARGS((const char *s,int c));
#endif /* STDC */

#if    defined(SHORTNAMES)
/*
 * Provide alternate external names which are unique in the first SIX
 * characters as required for portability (and Standard C)
 */
#define check_all	   chk_all
#define check_idx	   chk_idx
#define check_mixsym	   chk_mix
#define compare_number	   cmp_num
#define compare_one	   cmp_one
#define compare_page	   cmp_page
#define compare_string	   cmp_string
#define delim_n		   dlm_n
#define delim_p		   dlm_p
#define delim_r		   dlm_r
#define delim_t		   dlm_t
#define dump_sty	   dmpsty
#define encap_i		   ecp_i
#define encap_p		   ecp_p
#define encap_range	   ecp_range
#define encap_s		   ecp_s
#define flush_line	   flslin
#define flush_to_eol	   flseol
#define group_skip	   grp_skip
#define group_type	   grp_type
#define heading_pre	   hd_pre
#define heading_pre	   head_p
#define heading_suf	   hd_suf
#define heading_suf	   head_s
#define headings_flag	   hd_flag
#define headings_flag	   head_f
#define idx_aclose	   idxaclose
#define idx_actual	   idxactual
#define idx_keyword	   idx_kwd
#define indent_length	   ind_length
#define indent_space	   ind_space
#define numhead_neg	   num_neg
#define numhead_pos	   num_pos
#define process_idx	   prc_idx
#define process_precedence prc_pre
#define scan_alpha_lower   scnalw
#define scan_alpha_upper   scnaup
#define scan_arabic	   scnarabic
#define scan_arg1	   scna1
#define scan_arg2	   scna2
#define scan_char	   scnchr
#define scan_field	   scnfld
#define scan_idx	   scnidx
#define scan_key	   scnkey
#define scan_no		   scnno
#define scan_roman_lower   scnrlw
#define scan_roman_upper   scnrup
#define scan_spec	   scnspc
#define scan_string	   scnstr
#define scan_sty	   scnsty
#define setpage_close	   spg_close
#define setpage_open	   spg_open
#define setpagelen	   spg_len
#define suffix_2p	   suf_2p
#define suffix_3p	   suf_3p
#define suffix_mp	   suf_mp
#define symhead_neg	   sym_neg
#define symhead_pos	   sym_pos
#endif /* SHORTNAMES */

#ifndef	   R_OK
#define R_OK		4		/* only symbol from sys/file.h */
#endif

#undef TRUE
#define TRUE		1

#undef FALSE
#define FALSE		0

#undef NUL
#define NUL		'\0'

#undef NIL
#define NIL		""

#define TAB		'\t'
#define LFD		'\n'
#define SPC		' '
#define LSQ		'['
#define BSH		'\\'

#ifndef SW_PREFIX			/* can override at compile time */
#define SW_PREFIX	'-'
#endif

#define EXT_DELIM	'.'
#define EVEN		"even"
#define ODD		"odd"
#define ANY		"any"

#define GET_CHAR	getc

/* Revert to slow functions when fast macros unavailable for case conversion. */
/* tolower() is called more than any other function in MakeIndex. */

#ifndef _tolower
#define _tolower tolower
#endif

#ifndef _toupper
#define _toupper toupper
#endif

#if defined(__ACC__)
/* The macro forms do not work for Sun SunOS 4.1.x acc, force use of 
the functions */
#undef _tolower
#define _tolower tolower
#undef _toupper
#define _toupper toupper
#endif

#define ISLOWER(c)	islower(c)
#define ISUPPER(c)	isupper(c)

#if defined(__sgi)
/* Silicon Graphics ctype.h (all compilers) identifies characters 0300 .. 0337
as uppercase; we do NOT want this incorrect behavior. */
#undef ISLOWER
#define ISLOWER(c)	(((c) < 0200) && islower(c))
#undef ISUPPER
#define ISUPPER(c)	(((c) < 0200) && isupper(c))
#endif


/* NB: The typecasts here are CRITICAL for correct sorting of entries
   that use characters in the range 128..255! */

#define TOLOWER(C)	(ISUPPER((unsigned char)(C)) ? \
	(unsigned char)_tolower((unsigned char)(C)) : (unsigned char)(C))
#define TOUPPER(C)	(ISUPPER((unsigned char)(C)) ? \
	(unsigned char)(C) : (unsigned char)_toupper((unsigned char)(C)))

#define STREQ(A, B)	(strcmp(A, B) == 0)
#define STRNEQ(A, B)	(strcmp(A, B) != 0)

#define MESSAGE(F, S) { \
    if (verbose) \
	(void)fprintf(stderr, F, S); \
    (void)fprintf(ilg_fp, F, S); \
}

#define FATAL(F, S) { \
    (void)fprintf(stderr, F, S); \
    (void)fprintf(stderr, USAGE, pgm_fn); \
    EXIT(1); \
}

#define FATAL2(F, D1, D2) { \
    (void)fprintf(stderr, F, D1, D2); \
    (void)fprintf(stderr, USAGE, pgm_fn); \
    EXIT(1); \
}

#define CLOSE(FP)     fclose(FP)
#define OPEN_IN(FP)   fopen(FP, "r")
#define OPEN_OUT(FP)  fopen(FP, "w")
#define REOPEN_OUT(FN,FP) freopen(FN,"w",FP)

/*====================================================================
A few arrays in MakeIndex are dimensioned [xxx_MAX], where the xxx_MAX
values are defined below.  The use of each of these is described in
comments.  However, no run-time check is made to ensure that these are
consistent, or reasonable!  Therefore, change them only with great
care.

The array sizes should be made generously large: there are a great
many uses of strings in MakeIndex with the strxxx() and sprintf()
functions where no checking is done for adequate target storage sizes.
Although some input checking is done to avoid this possibility, it is
highly likely that there are still several places where storage
violations are possible, with core dumps, or worse, incorrect output,
ensuing.
======================================================================*/

#if !defined(PAGE_NUMBER_LONG) && !defined(PAGE_NUMBER_SHORT)
#define PAGE_NUMBER_LONG	1 /* default unless overridden by compiler */
#endif

#if defined(PAGE_NUMBER_SHORT)
typedef unsigned short page_number;
#if defined(USHRT_MAX)
#define PAGE_NUMBER_UNSET	USHRT_MAX
#else
#define PAGE_NUMBER_UNSET	((page_number)(-1))
#endif
#define ROMAN_LOWER_OFFSET 10000
#define ROMAN_UPPER_OFFSET 10000
#define ARABIC_OFFSET	   10000
#define ARABIC_MAX	4		/* number of digits in PAGE_MAX */
#define PAGE_MIN	0
#define PAGE_MAX     9999		/* one less than xxx_OFFSET */
#endif /* defined(PAGE_NUMBER_SHORT) */


#if defined(PAGE_NUMBER_LONG)
typedef unsigned long page_number;
#if defined(ULONG_MAX)
#define PAGE_NUMBER_UNSET	ULONG_MAX
#else
#define PAGE_NUMBER_UNSET	((page_number)(-1))
#endif
#define ROMAN_LOWER_OFFSET 100000000L
#define ROMAN_UPPER_OFFSET 100000000L
#define ARABIC_OFFSET	   100000000L
#define ARABIC_MAX	8		/* number of digits in PAGE_MAX */
#define PAGE_MIN	0L
#define PAGE_MAX 99999999L		/* one less than xxx_OFFSET */
#endif /* defined(PAGE_NUMBER_LONG) */


#define IN_PAGE_RANGE(n) ((PAGE_MIN <= (n)) && ((n) <= PAGE_MAX))
				/* Range of acceptable page numbers; */
				/* must be coordinated with ARABIC_MAX */

#define ARGUMENT_MAX  2000	/* maximum length of sort key + actual key */
				/* in index entry; static array storage is */
				/* 5*ARGUMENT_MAX. */

#define ARRAY_MAX     5000	/* maximum length of constant values in */
				/* style file; stack array storage is */
				/* three instances of ARRAY_MAX, allocated */
				/* statically on small machines. */
				/* MakeIndex 2.13 note: if you change */
				/* ARRAY_MAX, you must also change */
				/* references to it in ../doc/makeindex.l */
				/* as well. */

#define FIELD_MAX	3	/* maximum levels of index entries (item, */
				/* subitem, subsubitem); cannot be */
				/* increased beyond 3 without */
				/* significant design changes (no field */
				/* names are known beyond 3 levels) */

#ifndef FILENAME_MAX		/* Standard C has this in stdio.h */
#define	FILENAME_MAX    1024	/* max # of characters in a file name */
#endif

#if (FILENAME_MAX < 128)	/* guard against defective C implementations */
#undef FILENAME_MAX		/* which confuse file base name length with */
#define FILENAME_MAX	1024	/* file path name length */
#endif

#define LINELENGTH_MAX 72	/* maximum output line length (longer */
				/* ones wrap if possible); NOT used for */
				/* array sizes.  Can be overridden by */
				/* e.g. "line_max 127" in a style file. */

#ifndef NUMBER_MAX		/* allow compile-time definition */
#define NUMBER_MAX    100	/* one more than the maximum number of */
				/* digits in a (possibly composite) Roman */
				/* or Arabic page number, i.e. limited by */
				/* MAX(ARABIC_MAX,ROMAN_MAX).  From */
				/* MakeIndex 2.13 on, there is no */
				/* storage penalty for increasing this */
				/* value, since all index data strings */
				/* are dynamically allocated. */
#endif

#ifndef PAGEFIELD_MAX
#define PAGEFIELD_MAX (NUMBER_MAX/2)
				/* maximum fields in a composite page */
				/* number; don't increase this without */
				/* increasing NUMBER_MAX and ROMAN_MAX. */
				/* From MakeIndex 2.13 on, there is no */
				/* storage penalty for increasing this */
				/* value, since all index data strings */
				/* are dynamically allocated. */
#endif

#define PAGETYPE_MAX  5		/* fixed at 5; see use in scanst.c */

#define ROMAN_MAX     NUMBER_MAX /* maximum length of Roman page number */
				/* field; code in scanid.c assumes that */
				/* ROMAN_MAX == NUMBER_MAX */

/*====================================================================*/

#define VERSION	      "portable version 2.13 (beta) [11-Apr-1994]"

#if 0
#define VERSION	      "portable version 2.12 [26-May-1993]"
#define VERSION	      "portable version 2.11 [19-Oct-1991]"
#define VERSION	      "portable version 2.10 [05-Jul-1991]"
#define VERSION	      "portable version 2.9.1 [21-Jun-1990]"
#define VERSION	      "portable version 2.9 [13-Dec-1989]"
#define VERSION	      "portable version 2.8 [11-Nov-1989]"
#define VERSION	      "portable version 2.7 [01-Oct-1988]"
#define VERSION	      "portable version 2.6 [14-Jul-1988]"
#define VERSION	      "portable version 2.5 [14-Apr-1988]"
#define VERSION	      "portable version 2.4 [20-Mar-1988]"
#define VERSION	      "portable version 2.3 [20-Jan-1988]"
#define VERSION	      "version 2.2 [29-May-1987]"
#endif

#define PUT_VERSION { \
    MESSAGE("This is %s, ", pgm_fn); \
    MESSAGE("%s.\n", VERSION); \
    need_version = FALSE; \
}

#define USAGE \
   "Usage: %s [-c] [-d] [-e errfile] [-g] [-i] [-l] [-o ind] [-p num] \n\
\t\t[-q] [-r] [-s sty] [-t log] [idxfile0 idxfile1 ...]\n"

#define STYLE_PATH "INDEXSTYLE"	/* environment variable defining search */
				/* path for style files */
#define INDEX_IDX	".idx"
#define INDEX_ILG	".ilg"
#define INDEX_IND	".ind"
#define INDEX_STY	".mst"

#if !defined(INDEX_LOG)
#define INDEX_LOG	".log"
#endif

#define ROML		((page_type)0)	/* page number type codes */
#define ROMU		((page_type)1)
#define ARAB		((page_type)2)
#define ALPL		((page_type)3)
#define ALPU		((page_type)4)
#define ARABALPHA	((page_type)5)
#define EMPTY		((page_type)6)
#define DUPLICATE	((page_type)7)

#define SYMBOL		((page_group)(-1)) /* group type codes MUST be */
#define ALPHA		((page_group)(-2)) /* negative because of assumptions */
					/* in sortid.c:compar_one() and */
					/* scanid.c:group_type() */

#define GERMAN		0		/* argument to new_strcmp() in
					   sortid.c:compare_string() */

/***********************************************************************
Line numbers are recorded with every index entry.  They serve two
purposes: (1) to identify input file location in error messages, and (2)
to order identical page encap entries.  We therefore choose an unsigned
short, rather than an int, to represent them, because this saves 2 bytes
per index entry, and using unsigned doubles the range.  The largest
index file in the test suite is just over 25,000 lines long, and
produces a typeset 2-column index of over 300 pages.  Thus, there seems
to be no reason to sacrifice storage by using an int instead of a short.
Similar, entry numbers are essentially equivalent to line numbers, since
the .idx file format allows only one entry per line, so we can use the
same data type for them too.
***********************************************************************/

typedef unsigned short	line_number;
typedef unsigned short	entry_number;

/***********************************************************************

makeindex 2.13: Since we really never have to repeatedly allocate and
free dynamic string storage, we use chkalloc() to allocate it with
less storage overhead than malloc(), by using malloc() to get large
chunks within which we suballocate small chunks using just the storage
that they need, subject to alignment requirements.  At job
termination, we free the large chunks with chkfree(), so that there
are no untidy memory leaks.  All storage allocation in MakeIndex 2.13
or later uses chkalloc() instead of malloc().

We further reduce storage requirements by storing only the page
numbers that are actually used, in structures of type PAGE_PARTS.  In
order to support page number strings of mixed arabic and alphabetic
characters, such as 120A, 120B, ..., we store the original page number
string in the page[] array of PAGE_PARTS.  Although that structure
declares the page[] array of 1 character, don't believe that size.  We
actually allocate as many characters as are needed, in the ENTER()
macro in scanid.h.  The structure names in PAGE_PARTS merely provide
convenient handles to access the various fields.  Integer storage
alignment requirements will require allocating each PAGE_PARTS
structure on a suitable boundary, but that can be conveniently hidden
inside chkalloc().

***********************************************************************/

typedef unsigned char	page_type;
typedef unsigned char	page_count;

#if STDC
typedef signed char	page_group;		/* MUST be a signed type */
#else
	/* signed attribute not available in pre-Standard C environments, */
	/* and implementation dependent whether char is signed or unsigned */
typedef short		page_group;		/* MUST be a signed type */
#endif

typedef struct KPAGE_PARTS
{
    page_number		pno;		/* biased numerical page value */
    page_type		type;		/* type code (ROML .. ARABALPHA,
					   EMPTY, DUPLICATE) */
    char		page[1];	/* NUL-terminated string of 0 or */
					/* more non-NUL characters */
} PAGE_PARTS, *PAGE_PARTS_PTR;

/* Convenient access macros to conceal ugly pointers */
#define PAGE_VALUE(fieldptr,index)	(((fieldptr)->ppp[index])->pno)
#define PAGE_TYPE(fieldptr,index)	(((fieldptr)->ppp[index])->type)
#define PAGE_STRING(fieldptr,index)	(((fieldptr)->ppp[index])->page)

/* NB: FIELD is the most important data structure in MakeIndex,
because it is used for EVERY index entry.  Its contents and element
order and data types have been carefully chosen to minimize storage
requirements while still satisfying storage alignment restrictions of
many computer architectures.  If the typedefs for line_number,
page_group, or page_count are ever changed, it is IMPERATIVE to
reexamine this structure to make sure storage alignment requirements
are not violated, and space is not wasted. */

typedef struct KFIELD
{
    char		*sf[FIELD_MAX];	/* sort keys */
    char		*af[FIELD_MAX];	/* actual keys */
    char    		*lpg;		/* literal page number */
    PAGE_PARTS_PTR	*ppp;		/* page part pointer array */
    char		*encap;		/* encapsulator */
    const char		*fn;		/* input filename */
    line_number		lc;		/* line number in input .idx file */
    page_group		group;		/* key group */
    page_count		count;		/* page parts count */
}	FIELD, *FIELD_PTR;

typedef struct KNODE
{
    FIELD   data;
    struct KNODE *next;
}	NODE, *NODE_PTR;

extern FILE		*stddbg;

extern int		letter_ordering;
extern int		compress_blanks;
extern int		debug;
extern int		init_page;
extern int		merge_page;
extern int		even_odd;
extern int		verbose;
extern int		german_sort;

extern const char	*idx_keyword;
extern char		idx_aopen;
extern char		idx_aclose;
extern char		idx_level;
extern char		idx_ropen;
extern char		idx_rclose;
extern char		idx_quote;
extern char		idx_actual;
extern char		idx_encap;
extern char		idx_escape;

extern const char	*page_comp;
extern const char	*pgm_fn; /* used in FATAL*() and PUT_VERSION() */

extern page_number	page_offset[PAGETYPE_MAX];

extern const char	*preamble;
extern const char	*postamble;
extern const char	*setpage_open;
extern const char	*setpage_close;
extern const char	*group_skip;
extern int		headings_flag;
extern const char	*heading_pre;
extern const char	*heading_suf;
extern const char	*symhead_pos;
extern const char	*symhead_neg;
extern const char	*numhead_pos;
extern const char	*numhead_neg;
extern int		prelen;
extern int		postlen;
extern int		skiplen;
extern int		headprelen;
extern int		headsuflen;
extern int		setpagelen;

extern const char	*item_r[FIELD_MAX];
extern const char	*item_u[FIELD_MAX];
extern const char	*item_x[FIELD_MAX];
extern int		ilen_r[FIELD_MAX];
extern int		ilen_u[FIELD_MAX];
extern int		ilen_x[FIELD_MAX];

extern const char	*delim_p[FIELD_MAX];
extern const char	*delim_n;
extern const char	*delim_r;
extern const char	*delim_t;

extern const char	*suffix_2p;
extern const char	*suffix_3p;
extern const char	*suffix_mp;

extern const char	*encap_p;
extern const char	*encap_i;
extern const char	*encap_s;

extern int		linelength_max;
extern const char	*indent_space;
extern int		indent_length;

extern FILE		*idx_fp;
extern FILE		*sty_fp;
extern FILE		*ind_fp;
extern FILE		*ilg_fp;

extern const char	*idx_fn;
extern const char	*ind_fn;
extern const char	*ilg_fn;
extern const char	*sty_fn;

#ifndef MKIND_C
extern char		pageno[];
extern long		totmem;
#endif

extern FIELD_PTR	*idx_key;
extern NODE_PTR		head;
extern NODE_PTR		tail;

extern int		idx_dot;
extern line_number	idx_lc;
extern entry_number	idx_tt;
extern entry_number	idx_gt;
extern entry_number	idx_et;
extern int		idx_dc;

extern entry_number	sty_ec;

#define DOT		"."
#define DOT_MAX		1000 /* progress dot interval in gen_ind(), scan_idx()*/
#define CMP_MAX		1000 /* progress dot interval in sort stage */

/* Macro to print a dot every MAX invocations */
#define IDX_DOT(MAX) { \
    idx_dot = TRUE; \
    if (idx_dc++ == 0) { \
	if (verbose) \
	    (void)fprintf(stderr, DOT); \
	(void)fprintf(ilg_fp, DOT); \
    } \
    if (idx_dc == MAX) \
	idx_dc = 0; \
}

#define IDX_QUOTE_GERMAN '+'	/* use + instead of " in German style */

#define ALL_DONE { \
    if (fn_no > 0) { \
	if (verbose) \
(void)fprintf(stderr, \
	"Overall %d files read (%ld entries accepted, %ld rejected).\n", \
	(int)(fn_no+1), (long)idx_gt, (long)idx_et); \
(void)fprintf(ilg_fp, \
	"Overall %d files read (%ld entries accepted, %ld rejected).\n", \
	(int)(fn_no+1), (long)idx_gt, (long)idx_et); \
    } \
}

#define DONE(A, B, C, D) { \
    if (verbose) \
	(void)fprintf(stderr, "done (%ld %s, %ld %s).\n", \
		      (long)(A), (B), (long)(C), (D)); \
    (void)fprintf(ilg_fp, "done (%ld %s, %ld %s).\n", \
		  (long)(A), (B), (long)(C), (D)); \
}

extern void		*chkalloc ARGS((size_t size, size_t alignment_size));
extern void		chkfree ARGS((void));
extern int		compare_number ARGS((FIELD_PTR a,FIELD_PTR b,
					     page_count i));
extern void		dump_sty ARGS((void));
extern void		gen_ind ARGS((void));
extern int		group_type ARGS((char *str));
extern int		main ARGS((int argc, char **argv));
extern void		process_precedence ARGS((void));
extern void		qqsort ARGS((char *base, int n, int size,
				     int (*compar)ARGS((char*,char*))));
extern void		scan_idx ARGS((void));
extern void		scan_sty ARGS((void));
extern void		sort_idx ARGS((void));
extern page_number	strtopageno ARGS((char *str));
